/*----------------------------------------------------------------------
  File    : MLP.java
  Contents: training of very simple multi-layer perceptrons
  Author  : Christian Borgelt
  History : 2023.04.29 file created from file mlpd.c
----------------------------------------------------------------------*/
package jmlp;

import java.util.Random;
import java.io.IOException;
import java.io.File;
import java.io.FileWriter;
import java.io.FileReader;
import java.io.BufferedReader;
import java.io.Serializable;

/*--------------------------------------------------------------------*/
/** Class for simple multi-layer perceptron training.
 *  @author Christian Borgelt
 *  @since  2023.04.30 */
/*--------------------------------------------------------------------*/
public class MLP implements Serializable {

  /*------------------------------------------------------------------*/
  /*  constants                                                       */
  /*------------------------------------------------------------------*/
  private static final long serialVersionUID = 0x00010000;

  /*------------------------------------------------------------------*/
  /*  instance variables                                              */
  /*------------------------------------------------------------------*/
  /* --- neuron ids --- */
  /** the id of the first  hidden neuron */
  public static final int HD1    = 0;
  /** the id of the second hidden neuron */
  public static final int HD2    = 1;
  /** the id of the output neuron */
  public static final int OUT    = 2;

  /* --- connection ids --- */
  /** connection id: first  input */
  public static final int IN1    = 0;
  /** connection id: second input */
  public static final int IN2    = 1;
  /** connection id: (negated) bias */
  public static final int BIAS   = 2;

  /* --- problem types --- */
  /** problem type: exclusive or */
  public static final int XOR    = PatternSet.XOR;
  /** problem type: biimplication */
  public static final int BIIMP  = PatternSet.BIIMP;
  /** problem type: function 1 */
  public static final int FUNC1  = PatternSet.FUNC1;
  /** problem type: function 2 */
  public static final int FUNC2  = PatternSet.FUNC2;

  /* --- training mode --- */
  /** training mode: update after each pattern */
  public static final int ONLINE = 0;
  /** training mode: update after all patterns */
  public static final int BATCH  = 1; 

  /*------------------------------------------------------------------*/
  /*  instance variables                                              */
  /*------------------------------------------------------------------*/
  /** the connection weights */
  protected double[][] wgts   =
    { { 0, 0, 0 }, { 0, 0, 0 }, { 0, 0, 0 } };
  /** the connection weight changes */
  protected double[][] chgs   =
    { { 0, 0, 0 }, { 0, 0, 0 }, { 0, 0, 0 } };
  /** the neuron outputs */
  protected double[]   out    = { 0, 0, 0 };
  /** the problem type,  e.g. XOR */
  protected int        type   = XOR;
  /** the training mode, e.g. BATCH */
  protected int        mode   = BATCH;
  /** the learning rate */
  protected double     lrate  = 0.2;
  /** the momentum factor */
  protected double     moment = 0.0;
  /** the current training epoch */
  protected int        epoch  = 0;
  /** the current input pattern */
  protected int        curr   = 0;
  /** the training pattern set */
  protected PatternSet patset = new PatternSet(XOR);

  /*------------------------------------------------------------------*/
  /** Activation function (logistic function).
   *  @param  x the network input
   *  @return the computed activation
   *  @since  2023.04.30 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public static double act (double x)
  { return 1.0 /(1.0 +Math.exp(-(x))); }

  /*------------------------------------------------------------------*/
  /** Derivative of activation function (i.e. of logistic function).
   *  @param  x the activation value
   *  @return the derivative corresponding to the given activation
   *  @since  2023.04.30 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public static double deriv (double x)
  { return x *(1-x); }

  /*------------------------------------------------------------------*/
  /** Create a multi-layer perceptron.
   *  @param  type  the problem type identifier
   *  @since  2023.04.30 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public MLP (int type)
  {                             /* --- create an MLP */
    this.patset = new PatternSet(type);
    this.type   = this.patset.type();
  }  /* MLP() */

  /*------------------------------------------------------------------*/
  /** Set the problem type.
   *  @param  type  the problem type identifier
   *  @since  2023.04.30 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public void setType (int type)
  {                             /* --- set problem type */
    this.patset = new PatternSet(type);
    this.type   = this.patset.type();
  }  /* setType() */

  /*------------------------------------------------------------------*/
  /** Get the problem type of the multi-layer perceptron.
   *  (e.g. <code>XOR</code> or <code>FUNC1</code>)
   *  @return the problem type of the multi-layer perceptron
   *  @since  2023.04.30 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public int getType ()
  { return this.type; }

  /*------------------------------------------------------------------*/
  /** Set the training mode of the multi-layer perceptron.
   *  (<code>ONLINE</code> or <code>BATCH</code>).
   *  @param  mode  the training mode of the multi-layer perceptron
   *  @since  2023.04.30 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public void setMode (int mode)
  {                             /* --- set problem type */
    this.mode = (mode == ONLINE) ? ONLINE : BATCH;
  }  /* setMode() */

  /*------------------------------------------------------------------*/
  /** Set the training mode of the multi-layer perceptron.
   *  (<code>ONLINE</code> or <code>BATCH</code>).
   *  @return the training mode of the multi-layer perceptron
   *  @since  2023.04.30 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public int getMode ()
  { return this.mode; }

  /*------------------------------------------------------------------*/
  /** Set the learning rate of the multi-layer perceptron.
   *  @param  lrate  the learning rate
   *  @since  2023.04.30 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public void setLRate (double lrate)
  { this.lrate = lrate; }

  /*------------------------------------------------------------------*/
  /** Get the learning rate of the multi-layer perceptron.
   *  @return the learning rate of the multi-layer perceptron
   *  @since  2023.04.30 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public double getLRate ()
  { return this.lrate; }

  /*------------------------------------------------------------------*/
  /** Set the momentum factor of the multi-layer perceptron.
   *  @param  moment  the momentum factor of the multi-layer perceptron
   *  @since  2023.04.30 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public void setMoment (double moment)
  { this.moment = moment; }

  /*------------------------------------------------------------------*/
  /** Get the momentum factor of the multi-layer perceptron.
   *  @return the momentum factor of the multi-layer perceptron
   *  @since  2023.04.30 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public double getMoment ()
  { return this.moment; }

  /*------------------------------------------------------------------*/
  /** Set a connection weight or bias value.
   *  @param  nid  the neuron    identifier (<code>HD1</code>,
   *               <code>HD2</code> or <code>OUT</code>)
   *  @param  pid  the parameter identifier (<code>IN1</code>,
   *               <code>IN2</code> or <code>BIAS</code>)
   *  @param  w    the weight / bias value to set
   *  @since  2023.04.30 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public void setWeight (int nid, int pid, double w)
  { this.wgts[nid][pid] = w; }

  /*------------------------------------------------------------------*/
  /** Get a connection weight or bias value.
   *  @param  nid  the neuron    identifier (<code>HD1</code>,
   *               <code>HD2</code> or <code>OUT</code>)
   *  @param  pid  the parameter identifier (<code>IN1</code>,
   *               <code>IN2</code> or <code>BIAS</code>)
   *  @return the requested weight or bias value
   *  @since  2023.04.30 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public double getWeight (int nid, int pid)
  { return this.wgts[nid][pid]; }

  /*------------------------------------------------------------------*/
  /** Get the number of executed training epochs.
   *  @return the number of executed training epochs
   *  @since  2023.04.30 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public int getEpoch ()
  { return this.epoch; }

  /*------------------------------------------------------------------*/
  /** Get the output of a neuron.
   *  @param  nid  the neuron identifier
   *  @return the requested neuron output
   *  @since  2023.04.30 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public double getOutput (int nid)
  { return this.out[nid]; }

  /*------------------------------------------------------------------*/
  /** Load connection weights and bias values from a file.
   *  @param  file  the file to load from
   *  @throws IOException if a read error occurred
   *  @since  2023.04.30 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public void loadWeights (File file) throws IOException
  {                             /* --- save weights to file */
    BufferedReader reader = null;  /* reader for the file */
    try {                       /* open and read input file */
      reader = new BufferedReader(new FileReader(file));
      for (int i = 0; i < 3; i++) { /* traverse the neurons */
        String   line = reader.readLine();
        if (line == null)
          throw new IOException("wrong number of lines\n");
        String[] flds = line.split("\\s+");
        if (flds.length != 3)
          throw new IOException("wrong number of fields\n");
        this.wgts[i][0] = Double.parseDouble(flds[0]);
        this.wgts[i][1] = Double.parseDouble(flds[1]);
        this.wgts[i][2] = Double.parseDouble(flds[2]);
      } }                       /* read the weights */
    finally {                   /* close the input file */
      if (reader != null) reader.close(); }
  }  /* loadWeights() */

  /*------------------------------------------------------------------*/
  /** Save connection weights and bias values to a file.
   *  @param  file  the file to save to
   *  @throws IOException if a write error occurred
   *  @since  2023.04.30 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public void saveWeights (File file) throws IOException
  {                             /* --- save weights to file */
    FileWriter writer = null;   /* writer for the file */
    try {                       /* open and write to the file */
      writer = new FileWriter(file);
      for (int i = 0; i < 3; i++) { /* traverse the neurons */
        writer.write(String.format("%g ",  this.wgts[i][0]));
        writer.write(String.format("%g ",  this.wgts[i][1]));
        writer.write(String.format("%g\n", this.wgts[i][2]));
      } }                       /* write the weights */
    finally {                   /* close the output file */
      if (writer != null) writer.close(); }
  }  /* saveWeights() */

  /*------------------------------------------------------------------*/
  /** Get the training pattern set.
   *  @return the training pattern set
   *  @since  2023.04.30 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public PatternSet getPatternSet ()
  { return this.patset; }

  /*------------------------------------------------------------------*/
  /** Initialize the neural network.
   *  @param  range  the initial weight range
   *  @param  rand   the random number generator
   *  @since  2023.04.30 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public void init (double range, Random rand)
  {                             /* --- initialize neural network */
    if (rand != null) {         /* if to initialize the weights */
      for (int i = 3; --i >= 0;){ /* choose random connection weights */
        this.wgts[i][0] = 2*range *rand.nextDouble() -range;
        this.wgts[i][1] = 2*range *rand.nextDouble() -range;
        if (this.type >= FUNC1) /* if real-valued function */
          this.wgts[i][2] = 2*range *rand.nextDouble() -range;
        else {                  /* choose random weight, otherwise */
          this.wgts[i][2] = -rand.nextDouble() *this.wgts[i][0]
                            -rand.nextDouble() *this.wgts[i][1];
        }                       /* ensure that the line represented */
      }                         /* by each neuron is visible */
    }
    if (this.patset.inpcnt() < 2) /* if a real-valued function, clear */
      this.wgts[0][1] = this.wgts[1][1] = 0;  /* second input weights */
    this.chgs[0][0] = this.chgs[0][1] = this.chgs[0][2] = 0;
    this.chgs[1][0] = this.chgs[1][1] = this.chgs[1][2] = 0;
    this.chgs[2][0] = this.chgs[2][1] = this.chgs[2][2] = 0;
    this.epoch = 0;             /* clear weight changes and */
    this.curr  = 0;             /* init. epoch and current pattern */
  }  /* init() */

  /*------------------------------------------------------------------*/
  /** Calculate output of the neural network.
   *  @param  in1  the first  input value
   *  @param  in2  the second input value
   *  @return the output of the multi-layer perceptron
   *  @since  2023.04.30 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public double calc (double in1, double in2)
  {                               /* --- calculate network output */
    double x, y, z;               /* outputs of the three neurons */

    x = this.wgts[0][0] *in1 +this.wgts[0][1] *in2 +this.wgts[0][2];
    this.out[0] = x = act(x);
    y = this.wgts[1][0] *in1 +this.wgts[1][1] *in2 +this.wgts[1][2];
    this.out[1] = y = act(y);
    z = this.wgts[2][0] *x   +this.wgts[2][1] *y   +this.wgts[2][2];
    this.out[2] = z = act(z);     /* calculate and return */
    return z;                     /* the network output */
  }  /* calc() */

  /*------------------------------------------------------------------*/
  /** Train the neural network (one training step).
   *  @since  2023.04.30 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public void train ()
  {                             /* --- do one training step */
    int      last;              /* index of last training pattern */
    double   x,  y;             /* transformed coordinates */
    double   z,  ez;            /* network output and its error */
    double   ex, ey;            /* errors in hidden layer */
    double   d, t;              /* temporary buffers */
    double[] p;                 /* to traverse the training patterns */

    this.chgs[0][0] *= this.moment;
    this.chgs[0][1] *= this.moment;
    this.chgs[0][2] *= this.moment; /* first  input neuron */
    this.chgs[1][0] *= this.moment;
    this.chgs[1][1] *= this.moment;
    this.chgs[1][2] *= this.moment; /* second input neuron */
    this.chgs[2][0] *= this.moment;
    this.chgs[2][1] *= this.moment;
    this.chgs[2][2] *= this.moment; /* output neuron */
    if (this.curr >= this.patset.size()) {
      this.curr = 0; this.epoch++; }
    last = (this.mode == BATCH) /* get index of last pattern */
         ? this.patset.size()-1 : this.curr;
    for (; this.curr <= last; this.curr++) {
      p = this.patset.pattern(this.curr);
      t = this.wgts[0][0] *p[0] +this.wgts[0][1] *p[1] +this.wgts[0][2];
      x = act(t);
      t = this.wgts[1][0] *p[0] +this.wgts[1][1] *p[1] +this.wgts[1][2];
      y = act(t);
      t = this.wgts[2][0] *x    +this.wgts[2][1] *y    +this.wgts[2][2];
      z = act(t);               /* compute the network output */
      ez = p[2] -z;             /* and its error */
      this.chgs[2][2] += d = this.lrate *(t = ez *deriv(z));
      this.chgs[2][1] += d *y;
      this.chgs[2][0] += d *x;
      ey = t *this.wgts[2][1];  /* backpropagate error for y */
      this.chgs[1][2] += d = this.lrate *(ey *deriv(y));
      this.chgs[1][1] += d *p[1];
      this.chgs[1][0] += d *p[0];
      ex = t *this.wgts[2][0];  /* backpropagate error for x */
      this.chgs[0][2] += d = this.lrate *(ex *deriv(x));
      this.chgs[0][1] += d *p[1];
      this.chgs[0][0] += d *p[0];
    }
    this.wgts[2][2] += this.chgs[2][2];
    this.wgts[2][1] += this.chgs[2][1];
    this.wgts[2][0] += this.chgs[2][0]; /* update output neuron */
    this.wgts[1][2] += this.chgs[1][2];
    this.wgts[1][1] += this.chgs[1][1];
    this.wgts[1][0] += this.chgs[1][0]; /* update second input neuron */
    this.wgts[0][2] += this.chgs[0][2];
    this.wgts[0][1] += this.chgs[0][1];
    this.wgts[0][0] += this.chgs[0][0]; /* update first  input neuron */
  }  /* train() */

}  /* class MLP */
